// Public definitions for the POSIX-like file descriptor emulation layer
// that our user-land support library implements for the use of applications.
// See the code in the lib directory for the implementation details.

#ifndef JOS_INC_FD_H
#define JOS_INC_FD_H

#include <inc/types.h>
#include <inc/fs.h>

// Maximum number of file descriptors a program may hold open concurrently
#define MAXFD		32
// Bottom of file data area
#define FILEBASE	0xD0000000
// Bottom of file descriptor area
#define FDTABLE		(FILEBASE - PTSIZE)

// pre-declare for forward references
struct Dev;
struct Stat;

struct Fd {
	int fd_type;
	off_t fd_offset;
	int fd_omode;

	// Return the Fd object corresponding to file descriptor 'fdnum'.
	// Returns NULL if 'fdnum' is out of range.
	// Also returns NULL if 'require_open' is true and 'fdnum' is not open. 
	static Fd *fd(int fdnum, bool require_open);

	// Return a currently unused file descriptor.
	// Returns NULL if all fds are in use.
	static Fd *unused_fd();

	// Return the file descriptor number corresponding to this Fd.
	int fd_number() const {
		assert(round_down(this, PGSIZE) == this);
		assert(this >= (Fd *) FDTABLE);
		assert(this < (Fd *) FDTABLE + MAXFD * PGSIZE);
		return ((uintptr_t) this - FDTABLE) / PGSIZE;
	}

	// Return true iff this file descriptor is open.
	bool is_open() const;

	// Return the pointer into the file data area for this file descriptor.
	char *data() const {
		return (char *) (FILEBASE + PTSIZE * fd_number());
	}

	// Return the device implementing this file descriptor's operations.
	Dev *device() const;
};

struct FileFd : public Fd {
	int fd_fileid;
	struct File fd_file;
};	

struct Stat {
	char st_name[MAXNAMELEN];
	off_t st_size;
	int st_isdir;
	Dev *st_dev;
};


// Each different type of file descriptor (console, pipe, or file) is
// implemented by a different Dev (for "device").
struct Dev {
	int dev_id;
	const char *dev_name;

	Dev(int dev_id, const char *dev_name);
	
	virtual ssize_t read(Fd *fd, void *buf, size_t len, off_t offset);
	virtual ssize_t write(Fd *fd, const void *buf, size_t len, off_t offset);
	virtual int close(Fd *fd);
	virtual int stat(Fd *fd, struct Stat *stat);
	virtual int trunc(Fd *fd, off_t newsize);
};

#endif	// not JOS_INC_FD_H
